// Rose trees with functional updates:

sealed trait Tree[A]
case class Leaf[A](a: A) extends Tree[A]
case class Node[A](children: Tree[A]*) extends Tree[A]

sealed trait Context[+A]
case object Empty extends Context[Nothing]
case class Cons[A](
  left: List[Tree[A]], 
  up: Context[A],
  right: List[Tree[A]]) extends Context[A]

case class Loc[A](
  tree: Tree[A], 
  context: Context[A]) {

def this(t: Tree[A])= this(t, Empty)

def goLeft: Loc[A] = context match {
  case Cons(l::le, up, ri) =>
    Loc(l, Cons(le, up, tree::ri))
  case _ => sys.error("goLeft of first")
  }

def goRight: Loc[A] = context match {
  case Cons(le, up, r::ri) =>
    Loc(r, Cons(tree::le, up, ri))
  case _ => sys.error("goRight of last")
  }

def goUp: Loc[A] = context match {
  case Empty => sys.error("goUp of empty")
  case Cons(le, up, ri) =>
    Loc(Node((le.reverse ++ (tree::ri)):_*), up)
  }

def goDown: Loc[A] = tree match {
  case Leaf(_) => sys.error("goDown at leaf")
  case Node() => sys.error("goDown at empty")
  case Node(t, ts@ _*) => 
    Loc(t, Cons(List(), context, ts.toList))
  }

def insertLeft(t: Tree[A]): Loc[A] = context match {
  case Empty => sys.error("insertLeft at empty")
  case Cons(le, up, ri) => Loc(tree, Cons(t::le, up, ri))
  }

def insertRight(t: Tree[A]): Loc[A] = context match {
  case Empty => sys.error("insertRight at empty")
  case Cons(le, up, ri) => Loc(tree, Cons(le, up, t::ri))
  }

def insertDown(t: Tree[A]): Loc[A] = tree match {
  case Leaf(_) => sys.error("insertDown at leaf")
  case Node(ts @_*) => Loc(t, Cons(Nil, context, ts.toList))
  }

def update(t: Tree[A]): Loc[A] = Loc(t, context)

def delete: Loc[A] = context match {
  case Empty => Loc(Node(),Empty)
  case Cons(le, up, r::ri) => Loc(r, Cons(le, up, ri))
  case Cons(l::le, up, Nil) => Loc(l, Cons(le, up, Nil))
  case Cons(Nil, up, Nil) => Loc(Node(), up)
  }

def path(ps: List[Int]): Loc[A] = ps match {
  case Nil   => this
  case i::ps if i == 0 => goDown.path(ps)
  case i::ps if i > 0  => goLeft.path((i-1)::ps)
  }

}
