module SimpleEditor where

-- A very simple text editor

type Text   = [String]    
data Pos    = Pos { line :: Int, col :: Int} 
data Editor = Ed  { text   :: Text
                  , cursor :: Pos }

go_left :: Editor -> Editor
go_left Ed{text= t, cursor= c}
  | col c == 0 = error "At start of line"
  | otherwise = Ed{text= t, cursor=c{col= col c- 1}}

go_right :: Editor-> Editor
go_right Ed{text= t, cursor= c}
  | col c == length (t !! line c) = error "At end of line"
  | otherwise = Ed{text= t, cursor= c{col= col c+ 1}}

-- Problem: insert and delete are O(n) where n is the number of characters
-- in the text.

-- Insert a character to the right of the cursor
insert :: Editor-> String-> Editor
insert Ed{text= t, cursor= c} text = 
  let (as, bs) = splitAt (col c) (t !! line c)
  in  Ed{text= updateAt (line c) t (as ++ text++ bs), 
         cursor= c{col= col c+1}

-- Delete the character under the cursor
delete :: Editor-> Editor
delete Ed{text= t, cursor= c} =
  let (as, bs) = splitAt (col c) (t !! line c)      
      bs'      = if null bs then bs else tail bs
      c'       = if col c == length (t !! line c) && col c > 0
                 then col c-1 else col c
  in  Ed{text= updateAt (line c) t (as ++ bs'),
         cursor= c{col= c'}}

updateAt :: Int-> [a]-> a-> [a]
updateAt n as a = case splitAt n as of 
  (bs, [])    -> error "updateAt: list too short."
  (bs, _:cs) -> bs ++ a : cs
