-- Simple rose trees
data Tree a = Leaf a
            | Node [Tree a]
            deriving (Eq, Show)

-- Update, reference by name (substitution)
upd1 :: Eq a=> a-> Tree a-> Tree a-> Tree a
upd1 a nu (Node ts) = Node (map (upd1 a nu) ts)
upd1 a nu l@(Leaf b) | a == b     = nu
                     | otherwise  = l 

-- Update, reference by path
type Path = [Int]
upd2 :: Path-> Tree a-> Tree a-> Tree a
upd2 [] nu _ = nu 
upd2 (p:ps) _ (Leaf _) = error "Malformed path (at leaf, path too long)"
upd2 (p:ps) nu (Node ts)
   | p >= length ts = error "Malformed path (at node, index too large)"
   | otherwise = let (t1, t2)= splitAt p ts
                 in  Node $ t1++ upd2 ps nu (head t2) : tail t2


-- Examples
t = Node [ Leaf "-"
         , Node [Leaf "*", Leaf "a", Leaf "b"]
         , Node [Leaf "*", Leaf "c", Leaf "d"]
         ]

s= Node [Leaf "+", Leaf "x", Leaf "y"]
t1= upd1 "b" s t
t2= upd2 [1,2] s t

